@@ -29,6 +29,7 @@
 #include "boxed.h"
 #include "union.h"
 #include "gerror.h"
+#include <gjs/runtime.h>
 #include <gjs/gjs-module.h>
 #include <gjs/compat.h>
 
@@ -364,7 +365,7 @@ gjs_callback_closure(ffi_cif *cif,
 
 out:
     if (!success) {
-        gjs_log_exception (context, NULL);
+        gjs_log_exception (context);
 
         /* Fill in the result with some hopefully neutral value */
         g_callable_info_load_return_type(trampoline->info, &ret_type);
@@ -1223,12 +1224,12 @@ uninit_cached_function_data (Function *function)
 }
 
 static void
-function_finalize(JSContext *context,
-                  JSObject  *obj)
+function_finalize(JSFreeOp *fop,
+                  JSObject *obj)
 {
     Function *priv;
 
-    priv = priv_from_js(context, obj);
+    priv = JS_GetPrivate(obj);
     gjs_debug_lifecycle(GJS_DEBUG_GFUNCTION,
                         "finalize, obj %p priv %p", obj, priv);
     if (priv == NULL)
@@ -1242,15 +1243,15 @@ function_finalize(JSContext *context,
 
 static JSBool
 get_num_arguments (JSContext *context,
-                   JSObject  *obj,
-                   jsid       id,
+                   JSObject **obj,
+                   jsid      *id,
                    jsval     *vp)
 {
     int n_args, n_jsargs, i;
     jsval retval;
     Function *priv;
 
-    priv = priv_from_js(context, obj);
+    priv = priv_from_js(context, *obj);
 
     if (priv == NULL)
         return JS_FALSE;
@@ -1366,13 +1367,16 @@ static struct JSClass gjs_function_class = {
     JS_ConvertStub,
     function_finalize,
     NULL,
-    NULL,
     function_call,
-    NULL, NULL, NULL, NULL, NULL
+    NULL, NULL, NULL
 };
 
 static JSPropertySpec gjs_function_proto_props[] = {
-    { "length", 0, JSPROP_READONLY | JSPROP_PERMANENT | JSPROP_SHARED, (JSPropertyOp)get_num_arguments, NULL },
+    { "length", 0,
+      (JSPROP_READONLY | JSPROP_PERMANENT | JSPROP_SHARED),
+      JSOP_WRAPPER((JSPropertyOp)get_num_arguments),
+      JSOP_WRAPPER(JS_StrictPropertyStub)
+    },
     { NULL }
 };
 
@@ -1389,7 +1393,8 @@ init_cached_function_data (JSContext      *context,
                            GType           gtype,
                            GICallableInfo *info)
 {
-    guint8 i, n_args, array_length_pos;
+    guint8 i, n_args;
+    int array_length_pos;
     GError *error = NULL;
     GITypeInfo return_type;
     GIInfoType info_type;
@@ -1438,8 +1443,8 @@ init_cached_function_data (JSContext      *context,
         GIDirection direction;
         GIArgInfo arg_info;
         GITypeInfo type_info;
-        guint8 destroy = -1;
-        guint8 closure = -1;
+        int destroy = -1;
+        int closure = -1;
         GITypeTag type_tag;
 
         if (function->param_types[i] == PARAM_SKIPPED)
@@ -1538,18 +1543,21 @@ function_new(JSContext      *context,
     JSObject *function;
     JSObject *global;
     Function *priv;
+    JSBool found;
 
     /* put constructor for GIRepositoryFunction() in the global namespace */
     global = gjs_get_import_global(context);
 
-    if (!gjs_object_has_property(context, global, gjs_function_class.name)) {
+    if (!JS_HasProperty(context, global, gjs_function_class.name, &found))
+        return NULL;
+    if (!found) {
         JSObject *prototype;
         JSObject *parent_proto;
         jsval native_function;
 
         JS_GetProperty(context, global, "Function", &native_function);
         /* We take advantage from that fact that Function.__proto__ is Function.prototype */
-        parent_proto = JS_GetPrototype(context, JSVAL_TO_OBJECT(native_function));
+        parent_proto = JS_GetPrototype(JSVAL_TO_OBJECT(native_function));
 
         prototype = JS_InitClass(context, global,
                                  /* parent prototype JSObject* for
@@ -1574,9 +1582,7 @@ function_new(JSContext      *context,
                                  /* funcs of constructor, MyConstructor.myfunc() */
                                  NULL);
         if (prototype == NULL)
-            gjs_fatal("Can't init class %s", gjs_function_class.name);
-
-        g_assert(gjs_object_has_property(context, global, gjs_function_class.name));
+            g_error("Can't init class %s", gjs_function_class.name);
 
         gjs_debug(GJS_DEBUG_GFUNCTION, "Initialized class %s prototype %p",
                   gjs_function_class.name, prototype);
@@ -1594,7 +1600,7 @@ function_new(JSContext      *context,
     GJS_INC_COUNTER(function);
 
     g_assert(priv_from_js(context, function) == NULL);
-    JS_SetPrivate(context, function, priv);
+    JS_SetPrivate(function, priv);
 
     gjs_debug_lifecycle(GJS_DEBUG_GFUNCTION,
                         "function constructor, obj %p priv %p", function, priv);
