@@ -27,6 +27,7 @@
 
 #include <gjs/gjs-module.h>
 #include <gjs/compat.h>
+#include <gjs/runtime.h>
 #include "boxed.h"
 #include "enumeration.h"
 #include "repo.h"
@@ -61,6 +62,7 @@ GJS_NATIVE_CONSTRUCTOR_DECLARE(error)
     Error *priv;
     Error *proto_priv;
     JSObject *proto;
+    jsid message_name, code_name;
     jsval v_message, v_code;
     gchar *message;
 
@@ -77,13 +79,13 @@ GJS_NATIVE_CONSTRUCTOR_DECLARE(error)
     GJS_INC_COUNTER(gerror);
 
     g_assert(priv_from_js(context, object) == NULL);
-    JS_SetPrivate(context, object, priv);
+    JS_SetPrivate(object, priv);
 
     gjs_debug_lifecycle(GJS_DEBUG_GERROR,
                         "GError constructor, obj %p priv %p",
                         object, priv);
 
-    proto = JS_GetPrototype(context, object);
+    proto = JS_GetPrototype(object);
     gjs_debug_lifecycle(GJS_DEBUG_GERROR, "GError instance __proto__ is %p", proto);
 
     /* If we're the prototype, then post-construct we'll fill in priv->info.
@@ -101,17 +103,21 @@ GJS_NATIVE_CONSTRUCTOR_DECLARE(error)
     g_base_info_ref( (GIBaseInfo*) priv->info);
     priv->domain = proto_priv->domain;
 
-    if (!gjs_object_require_property (context, JSVAL_TO_OBJECT(argv[0]),
-                                      "GError constructor", "message", &v_message))
+    message_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                                GJS_STRING_MESSAGE);
+    code_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                             GJS_STRING_CODE);
+    if (!gjs_object_require_property(context, JSVAL_TO_OBJECT(argv[0]),
+                                     "GError constructor", message_name, &v_message))
         return JS_FALSE;
-    if (!gjs_object_require_property (context, JSVAL_TO_OBJECT(argv[0]),
-                                      "GError constructor", "code", &v_code))
+    if (!gjs_object_require_property(context, JSVAL_TO_OBJECT(argv[0]),
+                                     "GError constructor", code_name, &v_code))
         return JS_FALSE;
     if (!gjs_string_to_utf8 (context, v_message, &message))
         return JS_FALSE;
 
-    priv->gerror = g_error_new_literal (priv->domain, JSVAL_TO_INT(v_code),
-                                        message);
+    priv->gerror = g_error_new_literal(priv->domain, JSVAL_TO_INT(v_code),
+                                       message);
 
     g_free (message);
 
@@ -124,12 +130,12 @@ GJS_NATIVE_CONSTRUCTOR_DECLARE(error)
 }
 
 static void
-error_finalize(JSContext *context,
-               JSObject  *obj)
+error_finalize(JSFreeOp *fop,
+               JSObject *obj)
 {
     Error *priv;
 
-    priv = priv_from_js(context, obj);
+    priv = JS_GetPrivate(obj);
     gjs_debug_lifecycle(GJS_DEBUG_GERROR,
                         "finalize, obj %p priv %p", obj, priv);
     if (priv == NULL)
@@ -147,11 +153,11 @@ error_finalize(JSContext *context,
 }
 
 static JSBool
-error_get_domain(JSContext *context, JSObject *obj, jsid id, jsval *vp)
+error_get_domain(JSContext *context, JSObject **obj, jsid *id, jsval *vp)
 {
     Error *priv;
 
-    priv = priv_from_js(context, obj);
+    priv = priv_from_js(context, *obj);
 
     if (priv == NULL)
         return JS_FALSE;
@@ -161,11 +167,11 @@ error_get_domain(JSContext *context, JSObject *obj, jsid id, jsval *vp)
 }
 
 static JSBool
-error_get_message(JSContext *context, JSObject *obj, jsid id, jsval *vp)
+error_get_message(JSContext *context, JSObject **obj, jsid *id, jsval *vp)
 {
     Error *priv;
 
-    priv = priv_from_js(context, obj);
+    priv = priv_from_js(context, *obj);
 
     if (priv == NULL)
         return JS_FALSE;
@@ -180,11 +186,11 @@ error_get_message(JSContext *context, JSObject *obj, jsid id, jsval *vp)
 }
 
 static JSBool
-error_get_code(JSContext *context, JSObject *obj, jsid id, jsval *vp)
+error_get_code(JSContext *context, JSObject **obj, jsid *id, jsval *vp)
 {
     Error *priv;
 
-    priv = priv_from_js(context, obj);
+    priv = priv_from_js(context, *obj);
 
     if (priv == NULL)
         return JS_FALSE;
@@ -259,6 +265,7 @@ error_constructor_value_of(JSContext *context, unsigned argc, jsval *vp)
     jsval v_self, v_prototype;
     Error *priv;
     jsval v_out;
+    jsid prototype_name;
 
     v_self = JS_THIS(context, vp);
     if (!JSVAL_IS_OBJECT(v_self)) {
@@ -267,10 +274,12 @@ error_constructor_value_of(JSContext *context, unsigned argc, jsval *vp)
         return JS_FALSE;
     }
 
+    prototype_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                                  GJS_STRING_PROTOTYPE);
     if (!gjs_object_require_property(context,
                                      JSVAL_TO_OBJECT(v_self),
                                      "constructor",
-                                     "prototype",
+                                     prototype_name,
                                      &v_prototype))
         return JS_FALSE;
 
@@ -310,91 +319,49 @@ static struct JSClass gjs_error_class = {
     error_finalize,
     NULL,
     NULL,
-    NULL,
-    NULL, NULL, NULL, NULL, NULL
+    NULL, NULL, NULL
 };
 
 /* We need to shadow all fields of GError, to prevent calling the getter from GBoxed
    (which would trash memory accessing the instance private data) */
 static JSPropertySpec gjs_error_proto_props[] = {
-    { "domain", PROP_DOMAIN, GJS_MODULE_PROP_FLAGS | JSPROP_READONLY, error_get_domain, NULL },
-    { "code", PROP_CODE, GJS_MODULE_PROP_FLAGS | JSPROP_READONLY, error_get_code, NULL },
-    { "message", PROP_MESSAGE, GJS_MODULE_PROP_FLAGS | JSPROP_READONLY, error_get_message, NULL },
+    { "domain", PROP_DOMAIN,
+      GJS_MODULE_PROP_FLAGS | JSPROP_READONLY,
+      JSOP_WRAPPER((JSPropertyOp)error_get_domain),
+      JSOP_WRAPPER(JS_StrictPropertyStub)
+    },
+    { "code", PROP_CODE,
+      GJS_MODULE_PROP_FLAGS | JSPROP_READONLY,
+      JSOP_WRAPPER((JSPropertyOp)error_get_code),
+      JSOP_WRAPPER(JS_StrictPropertyStub)
+    },
+    { "message", PROP_MESSAGE,
+      GJS_MODULE_PROP_FLAGS | JSPROP_READONLY,
+      JSOP_WRAPPER((JSPropertyOp)error_get_message),
+      JSOP_WRAPPER(JS_StrictPropertyStub)
+    },
     { NULL }
 };
 
 static JSFunctionSpec gjs_error_proto_funcs[] = {
-    { "toString", error_to_string, 0, GJS_MODULE_PROP_FLAGS },
+    { "toString", JSOP_WRAPPER((JSNative)error_to_string), 0, GJS_MODULE_PROP_FLAGS },
     JS_FS_END
 };
 
 static JSFunctionSpec gjs_error_constructor_funcs[] = {
-    { "valueOf", error_constructor_value_of, 0, GJS_MODULE_PROP_FLAGS },
+    { "valueOf", JSOP_WRAPPER((JSNative)error_constructor_value_of), 0, GJS_MODULE_PROP_FLAGS },
     JS_FS_END
 };
 
-JSObject*
-gjs_lookup_error_constructor(JSContext    *context,
-                             GIEnumInfo  *info)
-{
-    JSObject *ns;
-    JSObject *constructor;
-
-    ns = gjs_lookup_namespace_object(context, (GIBaseInfo*) info);
-
-    if (ns == NULL)
-        return NULL;
-
-    constructor = NULL;
-    if (gjs_define_error_class(context, ns, info,
-                               &constructor, NULL))
-        return constructor;
-    else
-        return NULL;
-}
-
-JSObject*
-gjs_lookup_error_prototype(JSContext   *context,
-                           GIEnumInfo  *info)
-{
-    JSObject *ns;
-    JSObject *proto;
-
-    ns = gjs_lookup_namespace_object(context, (GIBaseInfo*) info);
-
-    if (ns == NULL)
-        return NULL;
-
-    proto = NULL;
-    if (gjs_define_error_class(context, ns, info, NULL, &proto))
-        return proto;
-    else
-        return NULL;
-}
-
-JSClass*
-gjs_lookup_error_class(JSContext    *context,
-                       GIEnumInfo   *info)
-{
-    JSObject *prototype;
-
-    prototype = gjs_lookup_error_prototype(context, info);
-
-    return JS_GET_CLASS(context, prototype);
-}
-
-JSBool
+void
 gjs_define_error_class(JSContext    *context,
                        JSObject     *in_object,
-                       GIEnumInfo   *info,
-                       JSObject    **constructor_p,
-                       JSObject    **prototype_p)
+                       GIEnumInfo   *info)
 {
     const char *constructor_name;
     GIBoxedInfo *glib_error_info;
     JSObject *prototype, *parent_proto;
     JSObject *constructor;
-    jsval value;
     Error *priv;
 
     /* See the comment in gjs_define_boxed_class() for an
@@ -404,34 +371,9 @@ gjs_define_error_class(JSContext    *context,
 
     constructor_name = g_base_info_get_name( (GIBaseInfo*) info);
 
-    if (gjs_object_get_property(context, in_object, constructor_name, &value)) {
-        JSObject *constructor;
-
-        if (!JSVAL_IS_OBJECT(value)) {
-            gjs_throw(context, "Existing property '%s' does not look like a constructor",
-                         constructor_name);
-            return JS_FALSE;
-        }
-
-        constructor = JSVAL_TO_OBJECT(value);
-
-        gjs_object_get_property(context, constructor, "prototype", &value);
-        if (!JSVAL_IS_OBJECT(value)) {
-            gjs_throw(context, "error %s prototype property does not appear to exist or has wrong type", constructor_name);
-            return JS_FALSE;
-        } else {
-            if (prototype_p)
-                *prototype_p = JSVAL_TO_OBJECT(value);
-            if (constructor_p)
-                *constructor_p = constructor;
-
-            return JS_TRUE;
-        }
-    }
-
     g_irepository_require(NULL, "GLib", "2.0", 0, NULL);
     glib_error_info = (GIBoxedInfo*) g_irepository_find_by_name(NULL, "GLib", "Error");
-    parent_proto = gjs_lookup_boxed_prototype(context, glib_error_info);
+    parent_proto = gjs_lookup_generic_prototype(context, glib_error_info);
     g_base_info_unref((GIBaseInfo*)glib_error_info);
 
     if (!gjs_init_class_dynamic(context, in_object,
@@ -450,8 +392,8 @@ gjs_define_error_class(JSContext    *context,
                                 &gjs_error_constructor_funcs[0],
                                 &prototype,
                                 &constructor)) {
-        gjs_log_exception(context, NULL);
-        gjs_fatal("Can't init class %s", constructor_name);
+        gjs_log_exception(context);
+        g_error("Can't init class %s", constructor_name);
     }
 
     GJS_INC_COUNTER(gerror);
@@ -460,20 +402,12 @@ gjs_define_error_class(JSContext    *context,
     g_base_info_ref( (GIBaseInfo*) priv->info);
     priv->domain = g_quark_from_string (g_enum_info_get_error_domain(priv->info));
 
-    JS_SetPrivate(context, prototype, priv);
+    JS_SetPrivate(prototype, priv);
 
     gjs_debug(GJS_DEBUG_GBOXED, "Defined class %s prototype is %p class %p in object %p",
-              constructor_name, prototype, JS_GET_CLASS(context, prototype), in_object);
+              constructor_name, prototype, JS_GetClass(prototype), in_object);
 
     gjs_define_enum_values(context, constructor, priv->info);
-
-    if (constructor_p)
-        *constructor_p = constructor;
-
-    if (prototype_p)
-        *prototype_p = prototype;
-
-    return JS_TRUE;
 }
 
 static GIEnumInfo *
@@ -509,48 +443,30 @@ static void
 define_error_properties(JSContext *context,
                         JSObject  *obj)
 {
-    JSStackFrame *frame;
-    JSScript *script;
-    jsbytecode *pc;
-    jsval v;
-    GString *stack;
-    const char *filename;
-    GjsContext *gjs_context;
+    jsid stack_name, filename_name, linenumber_name;
+    jsval stack, fileName, lineNumber;
 
-    /* find the JS frame that triggered the error */
-    frame = NULL;
-    while (JS_FrameIterator(context, &frame)) {
-        if (JS_IsScriptFrame(context, frame))
-            break;
-    }
-
-    /* someone called gjs_throw at top of the stack?
-       well, no stack in that case
-    */
-    if (!frame)
+    if (!gjs_context_get_frame_info (context,
+                                     &stack,
+                                     &fileName,
+                                     &lineNumber))
         return;
 
-    script = JS_GetFrameScript(context, frame);
-    pc = JS_GetFramePC(context, frame);
+    stack_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                              GJS_STRING_STACK);
+    filename_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                                 GJS_STRING_FILENAME);
+    linenumber_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                                   GJS_STRING_LINE_NUMBER);
 
-    stack = g_string_new(NULL);
-    gjs_context = JS_GetContextPrivate(context);
-    gjs_context_print_stack_to_buffer(gjs_context, frame, stack);
-
-    if (gjs_string_from_utf8(context, stack->str, stack->len, &v))
-        JS_DefineProperty(context, obj, "stack", v,
+    JS_DefinePropertyById(context, obj, stack_name, stack,
                           NULL, NULL, JSPROP_ENUMERATE);
 
-    filename = JS_GetScriptFilename(context, script);
-    if (gjs_string_from_filename(context, filename, -1, &v))
-        JS_DefineProperty(context, obj, "fileName", v,
+    JS_DefinePropertyById(context, obj, filename_name, fileName,
                           NULL, NULL, JSPROP_ENUMERATE);
 
-    v = INT_TO_JSVAL(JS_PCToLineNumber(context, script, pc));
-    JS_DefineProperty(context, obj, "lineNumber", v,
-                      NULL, NULL, JSPROP_ENUMERATE);
-
-    g_string_free(stack, TRUE);
+    JS_DefinePropertyById(context, obj, linenumber_name, lineNumber,
+                          NULL, NULL, JSPROP_ENUMERATE);
 }
 
 JSObject*
@@ -586,16 +502,16 @@ gjs_error_from_gerror(JSContext             *context,
                       "Wrapping struct %s %p with JSObject",
                       g_base_info_get_name((GIBaseInfo *)info), gboxed);
 
-    proto = gjs_lookup_error_prototype(context, info);
+    proto = gjs_lookup_generic_prototype(context, info);
     proto_priv = priv_from_js(context, proto);
 
     obj = JS_NewObjectWithGivenProto(context,
-                                     JS_GET_CLASS(context, proto), proto,
+                                     JS_GetClass(proto), proto,
                                      gjs_get_import_global (context));
 
     GJS_INC_COUNTER(gerror);
     priv = g_slice_new0(Error);
-    JS_SetPrivate(context, obj, priv);
+    JS_SetPrivate(obj, priv);
     priv->info = info;
     priv->domain = proto_priv->domain;
     g_base_info_ref( (GIBaseInfo*) priv->info);
