@@ -27,6 +27,7 @@
 
 #include <gjs/gjs-module.h>
 #include <gjs/compat.h>
+#include <gjs/runtime.h>
 
 #include <util/log.h>
 #include <util/glib.h>
@@ -81,14 +82,14 @@ child_free(void *data)
 GJS_NATIVE_CONSTRUCTOR_DEFINE_ABSTRACT(keep_alive)
 
 static void
-keep_alive_finalize(JSContext *context,
-                    JSObject  *obj)
+keep_alive_finalize(JSFreeOp *fop,
+                    JSObject *obj)
 {
     KeepAlive *priv;
     void *key;
     void *value;
 
-    priv = priv_from_js(context, obj);
+    priv = JS_GetPrivate(obj);
 
     gjs_debug_lifecycle(GJS_DEBUG_KEEP_ALIVE,
                         "keep_alive finalizing, obj %p priv %p", obj, priv);
@@ -133,7 +134,7 @@ keep_alive_trace(JSTracer *tracer,
 {
     KeepAlive *priv;
 
-    priv = JS_GetPrivate(tracer->context, obj);
+    priv = JS_GetPrivate(obj);
 
     if (priv == NULL) /* prototype */
         return;
@@ -150,7 +151,6 @@ keep_alive_trace(JSTracer *tracer,
  */
 static struct JSClass gjs_keep_alive_class = {
     "__private_GjsKeepAlive", /* means "new __private_GjsKeepAlive()" works */
-    JSCLASS_MARK_IS_TRACE | /* TraceOp not MarkOp */
     JSCLASS_HAS_PRIVATE,
     JS_PropertyStub,
     JS_PropertyStub,
@@ -164,10 +164,7 @@ static struct JSClass gjs_keep_alive_class = {
     NULL,
     NULL,
     NULL,
-    NULL,
-    NULL,
-    JS_CLASS_TRACE(keep_alive_trace),
-    NULL
+    keep_alive_trace,
 };
 
 static JSPropertySpec gjs_keep_alive_proto_props[] = {
@@ -184,6 +181,7 @@ gjs_keep_alive_new(JSContext *context)
     KeepAlive *priv;
     JSObject *keep_alive;
     JSObject *global;
+    JSBool found;
 
     /* This function creates an unattached KeepAlive object; following our
      * general strategy, we have a single KeepAlive class with a constructor
@@ -199,7 +197,9 @@ gjs_keep_alive_new(JSContext *context)
 
     g_assert(global != NULL);
 
-    if (!gjs_object_has_property(context, global, gjs_keep_alive_class.name)) {
+    if (!JS_HasProperty(context, global, gjs_keep_alive_class.name, &found))
+        return JS_FALSE;
+    if (!found) {
         JSObject *prototype;
 
         gjs_debug(GJS_DEBUG_KEEP_ALIVE,
@@ -229,9 +229,7 @@ gjs_keep_alive_new(JSContext *context)
                                  /* funcs of constructor, MyConstructor.myfunc() */
                                  NULL);
         if (prototype == NULL)
-            gjs_fatal("Can't init class %s", gjs_keep_alive_class.name);
-
-        g_assert(gjs_object_has_property(context, global, gjs_keep_alive_class.name));
+            g_error("Can't init class %s", gjs_keep_alive_class.name);
 
         gjs_debug(GJS_DEBUG_KEEP_ALIVE, "Initialized class %s prototype %p",
                   gjs_keep_alive_class.name, prototype);
@@ -243,15 +241,15 @@ gjs_keep_alive_new(JSContext *context)
 
     keep_alive = JS_NewObject(context, &gjs_keep_alive_class, NULL, global);
     if (keep_alive == NULL) {
-        gjs_log_exception(context, NULL);
-        gjs_fatal("Failed to create keep_alive object");
+        gjs_log_exception(context);
+        g_error("Failed to create keep_alive object");
     }
 
     priv = g_slice_new0(KeepAlive);
     priv->children = g_hash_table_new_full(child_hash, child_equal, NULL, child_free);
 
     g_assert(priv_from_js(context, keep_alive) == NULL);
-    JS_SetPrivate(context, keep_alive, priv);
+    JS_SetPrivate(keep_alive, priv);
 
     gjs_debug_lifecycle(GJS_DEBUG_KEEP_ALIVE,
                         "keep_alive constructor, obj %p priv %p", keep_alive, priv);
@@ -323,58 +321,34 @@ gjs_keep_alive_remove_child(JSContext         *context,
                         &child);
 }
 
-#define GLOBAL_KEEP_ALIVE_NAME "__gc_this_on_context_destroy"
-
 static JSObject*
-gjs_keep_alive_get_from_parent(JSContext *context,
-                               JSObject  *parent)
-{
-    jsval value;
-
-    gjs_object_get_property(context, parent, GLOBAL_KEEP_ALIVE_NAME, &value);
-
-    if (JSVAL_IS_OBJECT(value))
-        return JSVAL_TO_OBJECT(value);
-    else
-        return NULL;
-}
-
-JSObject*
-gjs_keep_alive_get_global(JSContext *context)
-{
-    return gjs_keep_alive_get_from_parent(context,
-                                          JS_GetGlobalObject(context));
-}
-
-static JSObject*
-gjs_keep_alive_create_in_parent(JSContext *context,
-                                JSObject  *parent)
+gjs_keep_alive_create(JSContext *context)
 {
     JSObject *keep_alive;
 
     JS_BeginRequest(context);
 
     keep_alive = gjs_keep_alive_new(context);
+    if (!keep_alive)
+        g_error("could not create keep_alive on global object, no memory?");
 
-    if (!JS_DefineProperty(context, parent,
-                           GLOBAL_KEEP_ALIVE_NAME,
-                           OBJECT_TO_JSVAL(keep_alive),
-                           NULL, NULL,
-                           /* No ENUMERATE since this is a hidden
-                            * implementation detail kind of property
-                            */
-                           JSPROP_READONLY | JSPROP_PERMANENT))
-        gjs_fatal("no memory to define keep_alive property");
+    gjs_set_global_slot(context, GJS_GLOBAL_SLOT_KEEP_ALIVE, OBJECT_TO_JSVAL(keep_alive));
 
     JS_EndRequest(context);
     return keep_alive;
 }
 
-static JSObject*
-gjs_keep_alive_create_in_global(JSContext *context)
+JSObject*
+gjs_keep_alive_get_global(JSContext *context)
 {
-    return gjs_keep_alive_create_in_parent(context,
-                                           JS_GetGlobalObject(context));
+    jsval keep_alive;
+
+    keep_alive = gjs_get_global_slot(context, GJS_GLOBAL_SLOT_KEEP_ALIVE);
+
+    if (G_LIKELY (JSVAL_IS_OBJECT(keep_alive)))
+        return JSVAL_TO_OBJECT(keep_alive);
+
+    return gjs_keep_alive_create(context);
 }
 
 void
@@ -389,12 +363,6 @@ gjs_keep_alive_add_global_child(JSContext         *context,
 
     keep_alive = gjs_keep_alive_get_global(context);
 
-    if (!keep_alive)
-        keep_alive = gjs_keep_alive_create_in_global(context);
-
-    if (!keep_alive)
-        gjs_fatal("could not create keep_alive on global object, no memory?");
-
     gjs_keep_alive_add_child(context,
                              keep_alive,
                              notify, child, data);
@@ -415,8 +383,8 @@ gjs_keep_alive_remove_global_child(JSContext         *context,
     keep_alive = gjs_keep_alive_get_global(context);
 
     if (!keep_alive)
-        gjs_fatal("no keep_alive property on the global object, have you "
-                  "previously added this child?");
+        g_error("no keep_alive property on the global object, have you "
+                "previously added this child?");
 
     gjs_keep_alive_remove_child(context,
                                 gjs_keep_alive_get_global(context),
@@ -424,28 +392,3 @@ gjs_keep_alive_remove_global_child(JSContext         *context,
 
     JS_EndRequest(context);
 }
-
-JSObject*
-gjs_keep_alive_get_for_import_global(JSContext *context)
-{
-    JSObject *global;
-    JSObject *keep_alive;
-
-    global = gjs_get_import_global(context);
-
-    g_assert(global != NULL);
-
-    JS_BeginRequest(context);
-
-    keep_alive = gjs_keep_alive_get_from_parent(context, global);
-
-    if (!keep_alive)
-        keep_alive = gjs_keep_alive_create_in_parent(context, global);
-
-    if (!keep_alive)
-        gjs_fatal("could not create keep_alive on global object, no memory?");
-
-    JS_EndRequest(context);
-
-    return keep_alive;
-}
