@@ -41,6 +41,7 @@
 #include <gjs/gjs-module.h>
 #include <gjs/compat.h>
 #include <gjs/type-module.h>
+#include <gjs/runtime.h>
 
 #include <util/log.h>
 #include <girepository.h>
@@ -81,21 +82,19 @@ typedef struct
 
 enum {
     PROP_0,
-    PROP_JS_CONTEXT,
-    PROP_JS_OBJECT,
     PROP_JS_HANDLED,
 };
 
+static GSList *object_init_list;
+
 static struct JSClass gjs_object_instance_class;
 static GThread *gjs_eval_thread;
 static volatile gint pending_idle_toggles;
 
 GJS_DEFINE_PRIV_FROM_JS(ObjectInstance, gjs_object_instance_class)
 
-static JSObject*       peek_js_obj  (JSContext *context,
-                                     GObject   *gobj);
-static void            set_js_obj   (JSContext *context,
-                                     GObject   *gobj,
+static JSObject*       peek_js_obj  (GObject   *gobj);
+static void            set_js_obj   (GObject   *gobj,
                                      JSObject  *obj);
 
 typedef enum {
@@ -104,16 +103,6 @@ typedef enum {
     VALUE_WAS_SET
 } ValueFromPropertyResult;
 
-static GQuark
-gjs_context_quark(void)
-{
-    static GQuark val = 0;
-    if (!val)
-        val = g_quark_from_static_string ("gjs::context");
-
-    return val;
-}
-
 static GQuark
 gjs_is_custom_type_quark (void)
 {
@@ -248,17 +237,17 @@ static inline ObjectInstance *
 proto_priv_from_js(JSContext *context,
                    JSObject  *obj)
 {
-    return priv_from_js(context, JS_GetPrototype(context, obj));
+    return priv_from_js(context, JS_GetPrototype(obj));
 }
 
 /* a hook on getting a property; set value_p to override property's value.
  * Return value is JS_FALSE on OOM/exception.
  */
 static JSBool
-object_instance_get_prop(JSContext *context,
-                         JSObject  *obj,
-                         jsid       id,
-                         jsval     *value_p)
+object_instance_get_prop(JSContext            *context,
+                         JSHandleObject        obj,
+                         JSHandleId            id,
+                         JSMutableHandleValue  value_p)
 {
     ObjectInstance *priv;
     char *name;
@@ -267,12 +256,12 @@ object_instance_get_prop(JSContext *context,
     GValue gvalue = { 0, };
     JSBool ret = JS_TRUE;
 
-    if (!gjs_get_string_id(context, id, &name))
+    if (!gjs_get_string_id(context, *id._, &name))
         return JS_TRUE; /* not resolved, but no error */
 
-    priv = priv_from_js(context, obj);
+    priv = priv_from_js(context, *obj._);
     gjs_debug_jsprop(GJS_DEBUG_GOBJECT,
-                     "Get prop '%s' hook obj %p priv %p", name, obj, priv);
+                     "Get prop '%s' hook obj %p priv %p", name, *obj._, priv);
 
     if (priv == NULL) {
         /* If we reach this point, either object_instance_new_resolve
@@ -308,7 +297,7 @@ object_instance_get_prop(JSContext *context,
     g_value_init(&gvalue, G_PARAM_SPEC_VALUE_TYPE(param));
     g_object_get_property(priv->gobj, param->name,
                           &gvalue);
-    if (!gjs_value_from_g_value(context, value_p, &gvalue)) {
+    if (!gjs_value_from_g_value(context, value_p._, &gvalue)) {
         g_value_unset(&gvalue);
         ret = JS_FALSE;
         goto out;
@@ -324,23 +313,23 @@ object_instance_get_prop(JSContext *context,
  * be set. Return value is JS_FALSE on OOM/exception.
  */
 static JSBool
-object_instance_set_prop(JSContext *context,
-                         JSObject  *obj,
-                         jsid       id,
-                         JSBool     strict,
-                         jsval     *value_p)
+object_instance_set_prop(JSContext            *context,
+                         JSHandleObject        obj,
+                         JSHandleId            id,
+                         JSBool                strict,
+                         JSMutableHandleValue  value_p)
 {
     ObjectInstance *priv;
     char *name;
     GParameter param = { NULL, { 0, }};
     JSBool ret = JS_TRUE;
 
-    if (!gjs_get_string_id(context, id, &name))
+    if (!gjs_get_string_id(context, *id._, &name))
         return JS_TRUE; /* not resolved, but no error */
 
-    priv = priv_from_js(context, obj);
+    priv = priv_from_js(context, *obj._);
     gjs_debug_jsprop(GJS_DEBUG_GOBJECT,
-                     "Set prop '%s' hook obj %p priv %p", name, obj, priv);
+                     "Set prop '%s' hook obj %p priv %p", name, *obj._, priv);
 
     if (priv == NULL) {
         /* see the comment in object_instance_get_prop() on this */
@@ -350,7 +339,7 @@ object_instance_set_prop(JSContext *context,
         goto out;
 
     switch (init_g_param_from_property(context, name,
-                                       *value_p,
+                                       *value_p._,
                                        G_TYPE_FROM_INSTANCE(priv->gobj),
                                        &param,
                                        FALSE /* constructing */)) {
@@ -497,8 +486,8 @@ object_instance_new_resolve_no_info(JSContext       *context,
  */
 static JSBool
 object_instance_new_resolve(JSContext *context,
-                            JSObject  *obj,
-                            jsid       id,
+                            JSObject **obj,
+                            jsid      *id,
                             unsigned   flags,
                             JSObject **objp)
 {
@@ -509,15 +498,15 @@ object_instance_new_resolve(JSContext *context,
 
     *objp = NULL;
 
-    if (!gjs_get_string_id(context, id, &name))
+    if (!gjs_get_string_id(context, *id, &name))
         return JS_TRUE; /* not resolved, but no error */
 
-    priv = priv_from_js(context, obj);
+    priv = priv_from_js(context, *obj);
 
     gjs_debug_jsprop(GJS_DEBUG_GOBJECT,
                      "Resolve prop '%s' hook obj %p priv %p (%s.%s) gobj %p %s",
                      name,
-                     obj,
+                     *obj,
                      priv,
                      priv && priv->info ? g_base_info_get_namespace (priv->info) : "",
                      priv && priv->info ? g_base_info_get_name (priv->info) : "",
@@ -546,7 +535,7 @@ object_instance_new_resolve(JSContext *context,
      * we need to look at exposing interfaces. Look up our interfaces through
      * GType data, and then hope that *those* are introspectable. */
     if (priv->info == NULL) {
-        ret = object_instance_new_resolve_no_info(context, obj, objp, priv, name);
+        ret = object_instance_new_resolve_no_info(context, *obj, objp, priv, name);
         goto out;
     }
 
@@ -578,8 +567,8 @@ object_instance_new_resolve(JSContext *context,
                 goto out;
             }
 
-            gjs_define_function(context, obj, priv->gtype, vfunc);
-            *objp = obj;
+            gjs_define_function(context, *obj, priv->gtype, vfunc);
+            *objp = *obj;
             g_base_info_unref((GIBaseInfo *)vfunc);
             ret = JS_TRUE;
             goto out;
@@ -610,7 +599,7 @@ object_instance_new_resolve(JSContext *context,
      * https://bugzilla.gnome.org/show_bug.cgi?id=632922
      */
     if (method_info == NULL) {
-        ret = object_instance_new_resolve_no_info(context, obj, objp,
+        ret = object_instance_new_resolve_no_info(context, *obj, objp,
                                                   priv, name);
         goto out;
     } else {
@@ -625,12 +614,12 @@ object_instance_new_resolve(JSContext *context,
                   g_base_info_get_namespace( (GIBaseInfo*) priv->info),
                   g_base_info_get_name( (GIBaseInfo*) priv->info));
 
-        if (gjs_define_function(context, obj, priv->gtype, method_info) == NULL) {
+        if (gjs_define_function(context, *obj, priv->gtype, method_info) == NULL) {
             g_base_info_unref( (GIBaseInfo*) method_info);
             goto out;
         }
 
-        *objp = obj; /* we defined the prop in obj */
+        *objp = *obj; /* we defined the prop in obj */
 
         g_base_info_unref( (GIBaseInfo*) method_info);
     }
@@ -678,28 +667,6 @@ object_instance_props_to_g_parameters(JSContext   *context,
     gparams = g_array_new(/* nul term */ FALSE, /* clear */ TRUE,
                           sizeof(GParameter));
 
-    /* For custom types we register, we need to set additional
-       properties for the JS context and JS object, so that we can retrieve
-       them inside the constructor, when handling construct properties
-       There is no other way to set those, as we need them before
-       g_object_newv returns.
-       We also need to ensure that these are the first properties set
-       (luckily g_object_newv preserves the order)
-    */
-    if (g_type_get_qdata(gtype, gjs_is_custom_type_quark())) {
-        GParameter gparam = { "js-context", { 0, } };
-
-        g_value_init(&gparam.value, G_TYPE_POINTER);
-        g_value_set_pointer(&gparam.value, context);
-
-        g_array_append_val(gparams, gparam);
-
-        gparam.name = "js-object";
-        g_value_set_pointer(&gparam.value, obj);
-
-        g_array_append_val(gparams, gparam);
-    }
-
     if (argc == 0 || JSVAL_IS_VOID(argv[0]))
         goto out;
 
@@ -725,14 +692,14 @@ object_instance_props_to_g_parameters(JSContext   *context,
         jsval value;
         GParameter gparam = { NULL, { 0, }};
 
-        if (!gjs_get_string_id(context, prop_id, &name))
-            goto free_array_and_fail;
-
-        if (!gjs_object_require_property(context, props, "property list", name, &value)) {
+        if (!gjs_object_require_property(context, props, "property list", prop_id, &value)) {
             g_free(name);
             goto free_array_and_fail;
         }
 
+        if (!gjs_get_string_id(context, prop_id, &name))
+            goto free_array_and_fail;
+
         switch (init_g_param_from_property(context, name,
                                            value,
                                            gtype,
@@ -866,7 +833,7 @@ handle_toggle_down(JSContext *context,
     ObjectInstance *priv;
     JSObject *obj;
 
-    obj = peek_js_obj(context, gobj);
+    obj = peek_js_obj(gobj);
 
     priv = priv_from_js(context, obj);
 
@@ -908,7 +875,7 @@ handle_toggle_up(JSContext *context,
     if (!gc_already_blocked)
         gjs_block_gc();
 
-    obj = peek_js_obj(context, gobj);
+    obj = peek_js_obj(gobj);
 
     if (!obj) {
         /* Object already GC'd */
@@ -926,7 +893,7 @@ handle_toggle_up(JSContext *context,
      */
     if (priv->keep_alive == NULL) {
         gjs_debug_lifecycle(GJS_DEBUG_GOBJECT, "Adding object to keep alive");
-        priv->keep_alive = gjs_keep_alive_get_for_import_global(context);
+        priv->keep_alive = gjs_keep_alive_get_global(context);
         gjs_keep_alive_add_child(context, priv->keep_alive,
                                  gobj_no_longer_kept_alive_func,
                                  obj,
@@ -1129,7 +1096,7 @@ init_object_private (JSContext *context,
     GJS_INC_COUNTER(object);
 
     g_assert(priv_from_js(context, object) == NULL);
-    JS_SetPrivate(context, object, priv);
+    JS_SetPrivate(object, priv);
 
     gjs_debug_lifecycle(GJS_DEBUG_GOBJECT,
                         "obj instance constructor, obj %p priv %p", object, priv);
@@ -1156,8 +1123,8 @@ associate_js_gobject (JSContext      *context,
     priv = priv_from_js(context, object);
     priv->gobj = gobj;
 
-    g_assert(peek_js_obj(context, gobj) == NULL);
-    set_js_obj(context, gobj, object);
+    g_assert(peek_js_obj(gobj) == NULL);
+    set_js_obj(gobj, object);
 
 #if DEBUG_DISPOSE
     g_object_weak_ref(gobj, wrapped_gobj_dispose_notify, object);
@@ -1174,7 +1141,7 @@ associate_js_gobject (JSContext      *context,
      * the wrapper to be garbage collected (and thus unref the
      * wrappee).
      */
-    priv->keep_alive = gjs_keep_alive_get_for_import_global(context);
+    priv->keep_alive = gjs_keep_alive_get_global(context);
     gjs_keep_alive_add_child(context,
                              priv->keep_alive,
                              gobj_no_longer_kept_alive_func,
@@ -1211,11 +1178,18 @@ object_instance_init (JSContext *context,
         return JS_FALSE;
     }
 
+    /* Mark this object in the construction stack, it
+       will be popped in gjs_object_custom_init() later
+       down.
+    */
+    if (g_type_get_qdata(gtype, gjs_is_custom_type_quark()))
+        object_init_list = g_slist_prepend(object_init_list, *object);
+
     gobj = g_object_newv(gtype, n_params, params);
 
     free_g_params(params, n_params);
 
-    old_jsobj = peek_js_obj(context, gobj);
+    old_jsobj = peek_js_obj(gobj);
     if (old_jsobj != NULL && old_jsobj != *object) {
         /* g_object_newv returned an object that's already tracked by a JS
          * object. Let's assume this is a singleton like IBus.IBus and return
@@ -1277,10 +1251,13 @@ GJS_NATIVE_CONSTRUCTOR_DECLARE(object_instance)
     JSBool ret;
     jsval initer;
     jsval rval;
+    jsid object_init_name;
 
     GJS_NATIVE_CONSTRUCTOR_PRELUDE(object_instance);
 
-    if (!gjs_object_require_property(context, object, "GObject instance", "_init", &initer))
+    object_init_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                                    GJS_STRING_GOBJECT_INIT);
+    if (!gjs_object_require_property(context, object, "GObject instance", object_init_name, &initer))
         return JS_FALSE;
 
     rval = JSVAL_VOID;
@@ -1317,7 +1294,7 @@ object_instance_trace(JSTracer *tracer,
     ObjectInstance *priv;
     GList *iter;
 
-    priv = JS_GetPrivate(tracer->context, obj);
+    priv = JS_GetPrivate(obj);
 
     for (iter = priv->signals; iter; iter = iter->next) {
         ConnectData *cd = iter->data;
@@ -1327,12 +1304,12 @@ object_instance_trace(JSTracer *tracer,
 }
 
 static void
-object_instance_finalize(JSContext *context,
+object_instance_finalize(JSFreeOp  *fop,
                          JSObject  *obj)
 {
     ObjectInstance *priv;
 
-    priv = priv_from_js(context, obj);
+    priv = JS_GetPrivate(obj);
     gjs_debug_lifecycle(GJS_DEBUG_GOBJECT,
                         "finalize obj %p priv %p gtype %s gobj %p", obj, priv,
                         (priv && priv->gobj) ?
@@ -1362,9 +1339,9 @@ object_instance_finalize(JSContext *context,
                     priv->info ? g_base_info_get_name((GIBaseInfo*) priv->info) : g_type_name(priv->gtype));
         }
 
-        set_js_obj(context, priv->gobj, NULL);
+        set_js_obj(priv->gobj, NULL);
         g_object_remove_toggle_ref(priv->gobj, wrapped_gobj_toggle_notify,
-                                   JS_GetRuntime(context));
+                                   fop->runtime);
         priv->gobj = NULL;
     }
 
@@ -1384,10 +1361,14 @@ object_instance_finalize(JSContext *context,
          * into jsapi, but we have to do this or the keep alive could
          * be finalized later and call gobj_no_longer_kept_alive_func.
          */
-        gjs_keep_alive_remove_child(context, priv->keep_alive,
-                                    gobj_no_longer_kept_alive_func,
-                                    obj,
-                                    priv);
+        {
+            JSContext *context = JS_NewContext(fop->runtime, 8192);
+            gjs_keep_alive_remove_child(context, priv->keep_alive,
+                                        gobj_no_longer_kept_alive_func,
+                                        obj,
+                                        priv);
+            JS_DestroyContext(context);
+        }
     }
 
     if (priv->info) {
@@ -1404,14 +1385,66 @@ object_instance_finalize(JSContext *context,
     g_slice_free(ObjectInstance, priv);
 }
 
-JSObject*
-gjs_lookup_object_prototype(JSContext    *context,
-                            GType         gtype)
+static JSObject *
+gjs_lookup_object_prototype_from_info(JSContext    *context,
+                                      GIObjectInfo *info,
+                                      GType         gtype)
 {
+    JSObject *in_object;
+    JSObject *constructor;
+    const char *constructor_name;
+    jsval value;
+
+    if (info) {
+        in_object = gjs_lookup_namespace_object(context, (GIBaseInfo*) info);
+        constructor_name = g_base_info_get_name((GIBaseInfo*) info);
+    } else {
+        in_object = gjs_lookup_private_namespace(context);
+        constructor_name = g_type_name(gtype);
+    }
+
+    if (G_UNLIKELY (!in_object))
+        return NULL;
+
+    if (!JS_GetProperty(context, in_object, constructor_name, &value))
+        return NULL;
+
+    if (JSVAL_IS_VOID(value)) {
+        /* In case we're looking for a private type, and we don't find it,
+           we need to define it first.
+        */
+        gjs_define_object_class(context, in_object, NULL, gtype, &constructor);
+    } else {
+        if (G_UNLIKELY (!JSVAL_IS_OBJECT(value) || JSVAL_IS_NULL(value)))
+            return NULL;
+
+        constructor = JSVAL_TO_OBJECT(value);
+    }
+
+    g_assert(constructor != NULL);
+
+    if (!gjs_object_get_property_const(context, constructor,
+                                       GJS_STRING_PROTOTYPE, &value))
+        return NULL;
+
+    if (G_UNLIKELY (!JSVAL_IS_OBJECT(value)))
+        return NULL;
+
+    return JSVAL_TO_OBJECT(value);
+}
+
+static JSObject *
+gjs_lookup_object_prototype(JSContext *context,
+                            GType      gtype)
+{
+    GIObjectInfo *info;
     JSObject *proto;
 
-    if (!gjs_define_object_class(context, NULL, gtype, NULL, &proto))
-        return NULL;
+    info = (GIObjectInfo*)g_irepository_find_by_gtype(g_irepository_get_default(), gtype);
+    proto = gjs_lookup_object_prototype_from_info(context, info, gtype);
+    if (info)
+        g_base_info_unref((GIBaseInfo*)info);
+
     return proto;
 }
 
@@ -1736,8 +1769,7 @@ to_string_func(JSContext *context,
 static struct JSClass gjs_object_instance_class = {
     "GObject_Object",
     JSCLASS_HAS_PRIVATE |
-    JSCLASS_NEW_RESOLVE |
-    JSCLASS_MARK_IS_TRACE,
+    JSCLASS_NEW_RESOLVE,
     JS_PropertyStub,
     JS_PropertyStub,
     object_instance_get_prop,
@@ -1750,10 +1782,8 @@ static struct JSClass gjs_object_instance_class = {
     NULL,
     NULL,
     NULL,
-    NULL,
-    NULL,
-    JS_CLASS_TRACE(object_instance_trace),
-    NULL,
+    object_instance_trace,
+    
 };
 
 static JSBool
@@ -1781,12 +1811,12 @@ static JSPropertySpec gjs_object_instance_proto_props[] = {
 };
 
 static JSFunctionSpec gjs_object_instance_proto_funcs[] = {
-    { "_init", (JSNative)init_func, 0, 0 },
-    { "connect", (JSNative)connect_func, 0, 0 },
-    { "connect_after", (JSNative)connect_after_func, 0, 0 },
-    { "disconnect", (JSNative)disconnect_func, 0, 0 },
-    { "emit", (JSNative)emit_func, 0, 0 },
-    { "toString", (JSNative)to_string_func, 0, 0 },
+    { "_init", JSOP_WRAPPER((JSNative)init_func), 0, 0 },
+    { "connect", JSOP_WRAPPER((JSNative)connect_func), 0, 0 },
+    { "connect_after", JSOP_WRAPPER((JSNative)connect_after_func), 0, 0 },
+    { "disconnect", JSOP_WRAPPER((JSNative)disconnect_func), 0, 0 },
+    { "emit", JSOP_WRAPPER((JSNative)emit_func), 0, 0 },
+    { "toString", JSOP_WRAPPER((JSNative)to_string_func), 0, 0 },
     { NULL }
 };
 
@@ -1825,12 +1855,12 @@ gjs_define_static_methods(JSContext    *context,
     return JS_TRUE;
 }
 
-JSBool
-gjs_define_object_class(JSContext     *context,
-                        JSObject      *in_object,
-                        GType          gtype,
-                        JSObject     **constructor_p,
-                        JSObject     **prototype_p)
+void
+gjs_define_object_class(JSContext      *context,
+                        JSObject       *in_object,
+                        GIObjectInfo   *info,
+                        GType           gtype,
+                        JSObject      **constructor_p)
 {
     const char *constructor_name;
     JSObject *prototype;
@@ -1838,25 +1868,12 @@ gjs_define_object_class(JSContext     *context,
     JSObject *parent_proto;
     jsval value;
     ObjectInstance *priv;
-    GIObjectInfo *info = NULL;
     const char *ns;
+    GType parent_type;
 
+    g_assert(in_object != NULL);
     g_assert(gtype != G_TYPE_INVALID);
 
-    info = (GIObjectInfo*)g_irepository_find_by_gtype(g_irepository_get_default(), gtype);
-
-    if (!in_object) {
-        if (info)
-            in_object = gjs_lookup_namespace_object(context, (GIBaseInfo*) info);
-        else
-            in_object = gjs_lookup_private_namespace(context);
-
-        if (!in_object) {
-            if (info)
-                g_base_info_unref((GIBaseInfo*)info);
-            return FALSE;
-        }
-    }
     /*   http://egachine.berlios.de/embedding-sm-best-practice/apa.html
      *   http://www.sitepoint.com/blogs/2006/01/17/javascript-inheritance/
      *   http://www.cs.rit.edu/~atk/JavaScript/manuals/jsobj/
@@ -1897,73 +1914,22 @@ gjs_define_object_class(JSContext     *context,
      * JavaScript is SO AWESOME
      */
 
-    /* 'gtype' is the GType of a concrete class (if any) which may or may not
-     * be defined in the GIRepository. 'info' corresponds to the first known
-     * ancestor of 'gtype' (or the gtype itself.)
-     *
-     * For example:
-     * gtype=GtkWindow  info=Gtk.Window     (defined)
-     * gtype=GLocalFile info=GLib.Object    (not defined)
-     * gtype=GHalMount  info=GLib.Object    (not defined)
-     *
-     * Each GType needs to have distinct JS class, otherwise the JS class for
-     * first common parent in GIRepository gets used with conflicting gtypes
-     * when resolving GTypeInterface methods.
-     *
-     * In case 'gtype' is not defined in GIRepository use the type name as
-     * constructor assuming it is unique enough instead of sharing
-     * 'Object' (or whatever the first known ancestor is)
-     *
-     */
-    if (!info) {
-        constructor_name = g_type_name(gtype);
-    } else {
-        constructor_name = g_base_info_get_name((GIBaseInfo*) info);
-    }
-
-    if (gjs_object_get_property(context, in_object, constructor_name, &value)) {
-
-        if (!JSVAL_IS_OBJECT(value)) {
-            gjs_throw(context, "Existing property '%s' does not look like a constructor",
-                         constructor_name);
-            g_base_info_unref((GIBaseInfo*)info);
-            return FALSE;
-        }
-
-        constructor = JSVAL_TO_OBJECT(value);
-
-        gjs_object_get_property(context, constructor, "prototype", &value);
-        if (!JSVAL_IS_OBJECT(value)) {
-            gjs_throw(context, "prototype property does not appear to exist or has wrong type");
-            g_base_info_unref((GIBaseInfo*)info);
-            return FALSE;
-        } else {
-            if (prototype_p)
-                *prototype_p = JSVAL_TO_OBJECT(value);
-            if (constructor_p)
-                *constructor_p = constructor;
-
-            if (info)
-                g_base_info_unref((GIBaseInfo*)info);
-            return TRUE;
-        }
-    }
-
     parent_proto = NULL;
-    if (g_type_parent(gtype) != G_TYPE_INVALID) {
-       GType parent_gtype;
+    parent_type = g_type_parent(gtype);
+    if (parent_type != G_TYPE_INVALID)
+       parent_proto = gjs_lookup_object_prototype(context, parent_type);
 
-       parent_gtype = g_type_parent(gtype);
-       parent_proto = gjs_lookup_object_prototype(context, parent_gtype);
-    }
-
-    /* This is only used to disambiguate classes in the import global.
+    /* ns is only used to set the JSClass->name field (exposed by Object.prototype.toString).
      * We can safely set "unknown" if there is no info, as in that case
      * the name is globally unique (it's a GType name). */
-    if (info)
+    if (info) {
         ns = g_base_info_get_namespace((GIBaseInfo*) info);
-    else
+        constructor_name = g_base_info_get_name((GIBaseInfo*) info);
+    } else {
         ns = "unknown";
+        constructor_name = g_type_name(gtype);
+    }
+
     if (!gjs_init_class_dynamic(context, in_object,
                                 parent_proto,
                                 ns, constructor_name,
@@ -1979,20 +1945,20 @@ gjs_define_object_class(JSContext     *context,
                                 NULL,
                                 &prototype,
                                 &constructor)) {
-        gjs_fatal("Can't init class %s", constructor_name);
+        g_error("Can't init class %s", constructor_name);
     }
 
     GJS_INC_COUNTER(object);
     priv = g_slice_new0(ObjectInstance);
     priv->info = info;
     if (info)
-        g_base_info_ref( (GIBaseInfo*) priv->info);
+        g_base_info_ref((GIBaseInfo*) info);
     priv->gtype = gtype;
     priv->klass = g_type_class_ref (gtype);
-    JS_SetPrivate(context, prototype, priv);
+    JS_SetPrivate(prototype, priv);
 
     gjs_debug(GJS_DEBUG_GOBJECT, "Defined class %s prototype %p class %p in object %p",
-              constructor_name, prototype, JS_GET_CLASS(context, prototype), in_object);
+              constructor_name, prototype, JS_GetClass(prototype), in_object);
 
     if (info)
         gjs_define_static_methods(context, constructor, gtype, info);
@@ -2001,28 +1967,19 @@ gjs_define_object_class(JSContext     *context,
     JS_DefineProperty(context, constructor, "$gtype", value,
                       NULL, NULL, JSPROP_PERMANENT);
 
-    if (prototype_p)
-        *prototype_p = prototype;
-
     if (constructor_p)
         *constructor_p = constructor;
-
-    if (info)
-        g_base_info_unref((GIBaseInfo*)info);
-    return TRUE;
 }
 
 static JSObject*
-peek_js_obj(JSContext *context,
-            GObject   *gobj)
+peek_js_obj(GObject *gobj)
 {
     return g_object_get_qdata(gobj, gjs_object_priv_quark());
 }
 
 static void
-set_js_obj(JSContext *context,
-           GObject   *gobj,
-           JSObject  *obj)
+set_js_obj(GObject  *gobj,
+           JSObject *obj)
 {
     g_object_set_qdata(gobj, gjs_object_priv_quark(), obj);
 }
@@ -2036,24 +1993,24 @@ gjs_object_from_g_object(JSContext    *context,
     if (gobj == NULL)
         return NULL;
 
-    obj = peek_js_obj(context, gobj);
+    obj = peek_js_obj(gobj);
 
     if (obj == NULL) {
         /* We have to create a wrapper */
         JSObject *proto;
+        GType gtype;
 
         gjs_debug_marshal(GJS_DEBUG_GOBJECT,
                           "Wrapping %s with JSObject",
                           g_type_name_from_instance((GTypeInstance*) gobj));
 
-
-        if (!gjs_define_object_class(context, NULL, G_TYPE_FROM_INSTANCE(gobj), NULL, &proto))
-            return NULL;
+        gtype = G_TYPE_FROM_INSTANCE(gobj);
+        proto = gjs_lookup_object_prototype(context, gtype);
 
         JS_BeginRequest(context);
 
         obj = JS_NewObjectWithGivenProto(context,
-                                         JS_GET_CLASS(context, proto), proto,
+                                         JS_GetClass(proto), proto,
                                          gjs_get_import_global (context));
 
         JS_EndRequest(context);
@@ -2069,7 +2026,7 @@ gjs_object_from_g_object(JSContext    *context,
         /* see the comment in init_object_instance() for this */
         g_object_unref(gobj);
 
-        g_assert(peek_js_obj(context, gobj) == obj);
+        g_assert(peek_js_obj(gobj) == obj);
     }
 
  out:
@@ -2352,18 +2309,16 @@ gjs_object_get_gproperty (GObject    *object,
                           GValue     *value,
                           GParamSpec *pspec)
 {
+    GjsContext *gjs_context;
     JSContext *context;
     JSObject *js_obj;
     jsval jsvalue;
     gchar *underscore_name;
 
-    if (property_id != PROP_JS_HANDLED) {
-        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
-        return;
-    }
+    gjs_context = gjs_context_get_current();
+    context = gjs_context_get_native_context(gjs_context);
 
-    context = g_object_get_qdata(object, gjs_context_quark());
-    js_obj = peek_js_obj(context, object);
+    js_obj = peek_js_obj(object);
 
     underscore_name = hyphen_to_underscore((gchar *)pspec->name);
     JS_GetProperty(context, js_obj, underscore_name, &jsvalue);
@@ -2379,31 +2334,16 @@ gjs_object_set_gproperty (GObject      *object,
                           const GValue *value,
                           GParamSpec   *pspec)
 {
+    GjsContext *gjs_context;
     JSContext *context;
     JSObject *js_obj;
     jsval jsvalue;
     gchar *underscore_name;
 
-    if (property_id == PROP_JS_CONTEXT) {
-        context = g_value_get_pointer (value);
-        g_object_set_qdata(object, gjs_context_quark(), context);
-        return;
-    }
+    gjs_context = gjs_context_get_current();
+    context = gjs_context_get_native_context(gjs_context);
 
-    context = g_object_get_qdata(object, gjs_context_quark());
-
-    if (property_id == PROP_JS_OBJECT) {
-        js_obj = g_value_get_pointer (value);
-        associate_js_gobject(context, js_obj, object);
-        return;
-    }
-
-    if (property_id != PROP_JS_HANDLED) {
-        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
-        return;
-    }
-
-    js_obj = peek_js_obj(context, object);
+    js_obj = peek_js_obj(object);
 
     if (!gjs_value_from_g_value(context, &jsvalue, value))
         return;
@@ -2420,20 +2360,37 @@ gjs_object_class_init(GObjectClass *class,
     class->set_property = gjs_object_set_gproperty;
     class->get_property = gjs_object_get_gproperty;
 
-    g_object_class_install_property (class, PROP_JS_CONTEXT,
-                                     g_param_spec_pointer ("js-context",
-                                                           "JSContext",
-                                                           "The JSContext this object was created for",
-                                                           G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));
-
-    g_object_class_install_property (class, PROP_JS_OBJECT,
-                                     g_param_spec_pointer ("js-object",
-                                                           "JSObject",
-                                                           "The JSObject wrapping this GObject",
-                                                           G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));
     gjs_eval_thread = g_thread_self();
 }
 
+static void
+gjs_object_custom_init(GTypeInstance *instance,
+                       gpointer       g_class)
+{
+    GjsContext *gjs_context;
+    JSContext *context;
+    JSObject *object;
+    ObjectInstance *priv;
+
+    object = object_init_list->data;
+    priv = JS_GetPrivate(object);
+
+    if (priv->gtype != G_TYPE_FROM_INSTANCE (instance)) {
+        /* This is not the most derived instance_init function,
+           do nothing.
+         */
+        return;
+    }
+
+    object_init_list = g_slist_delete_link(object_init_list,
+                                           object_init_list);
+
+    gjs_context = gjs_context_get_current();
+    context = gjs_context_get_native_context(gjs_context);
+
+    associate_js_gobject(context, object, G_OBJECT (instance));
+}
+
 static inline void
 gjs_add_interface(GType instance_type,
                   GType interface_type)
@@ -2452,7 +2409,7 @@ gjs_register_type(JSContext *cx,
 {
     jsval *argv = JS_ARGV(cx, vp);
     gchar *name;
-    JSObject *parent, *constructor, *interfaces;
+    JSObject *parent, *constructor, *interfaces, *module;
     GType instance_type, parent_type;
     GTypeQuery query;
     GTypeModule *type_module;
@@ -2469,7 +2426,7 @@ gjs_register_type(JSContext *cx,
 
 	0,    /* instance_size */
 	0,    /* n_preallocs */
-	(GInstanceInitFunc) NULL,
+	gjs_object_custom_init,
     };
     guint32 i, n_interfaces;
     GType *iface_types;
@@ -2555,8 +2512,8 @@ gjs_register_type(JSContext *cx,
         gjs_add_interface(instance_type, iface_types[i]);
 
     /* create a custom JSClass */
-    if (!gjs_define_object_class(cx, NULL, instance_type, &constructor, NULL))
-        goto out;
+    module = gjs_lookup_private_namespace(cx);
+    gjs_define_object_class(cx, module, NULL, instance_type, &constructor);
 
     JS_SET_RVAL(cx, vp, OBJECT_TO_JSVAL(constructor));
 
