@@ -37,6 +37,7 @@
 #include "gerror.h"
 #include <gjs/gjs-module.h>
 #include <gjs/compat.h>
+#include <gjs/runtime.h>
 
 #include <girepository.h>
 
@@ -83,6 +84,9 @@ closure_marshal(GClosure        *closure,
 
         gjs_set_values(context, argv, argc, JSVAL_VOID);
         gjs_root_value_locations(context, argv, argc);
+    } else {
+        /* squash a compiler warning */
+        argv = NULL;
     }
     JS_AddValueRoot(context, &rval);
 
@@ -121,7 +125,7 @@ closure_marshal(GClosure        *closure,
             gjs_debug(GJS_DEBUG_GCLOSURE,
                       "Unable to convert arg %d in order to invoke closure",
                       i);
-            gjs_log_exception(context, NULL);
+            gjs_log_exception(context);
             goto cleanup;
         }
     }
@@ -137,7 +141,7 @@ closure_marshal(GClosure        *closure,
         if (!gjs_value_to_g_value(context, rval, return_value)) {
             gjs_debug(GJS_DEBUG_GCLOSURE,
                       "Unable to convert return value when invoking closure");
-            gjs_log_exception(context, NULL);
+            gjs_log_exception(context);
             goto cleanup;
         }
     }
@@ -352,17 +356,21 @@ gjs_value_to_g_value_internal(JSContext    *context,
 
         g_value_set_object(gvalue, gobj);
     } else if (gtype == G_TYPE_STRV) {
+        jsid length_name;
+        JSBool found_length;
+
+        length_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                                   GJS_STRING_LENGTH);
         if (JSVAL_IS_NULL(value)) {
             /* do nothing */
-        } else if (gjs_object_has_property(context,
-                                           JSVAL_TO_OBJECT(value),
-                                           "length")) {
+        } else if (JS_HasPropertyById(context, JSVAL_TO_OBJECT(value), length_name, &found_length) &&
+                   found_length) {
             jsval length_value;
             guint32 length;
 
             if (!gjs_object_require_property(context,
                                              JSVAL_TO_OBJECT(value), NULL,
-                                             "length",
+                                             length_name,
                                              &length_value) ||
                 !JS_ValueToECMAUint32(context, length_value, &length)) {
                 gjs_throw(context,
