@@ -30,6 +30,7 @@
 #include "native.h"
 #include "byteArray.h"
 #include "compat.h"
+#include "runtime.h"
 
 #include "gi.h"
 #include "gi/object.h"
@@ -57,8 +58,8 @@ static void     gjs_context_set_property      (GObject               *object,
                                                   guint                  prop_id,
                                                   const GValue          *value,
                                                   GParamSpec            *pspec);
-static JSBool gjs_on_context_gc (JSContext *cx,
-                                 JSGCStatus status);
+static void gjs_on_context_gc (JSRuntime *rt,
+                               JSGCStatus status);
 
 struct _GjsContext {
     GObject parent;
@@ -70,6 +71,7 @@ struct _GjsContext {
     GjsProfiler *profiler;
 
     char *jsversion_string;
+    char *program_name;
 
     char **search_path;
 
@@ -95,7 +97,8 @@ enum {
     PROP_0,
     PROP_JS_VERSION,
     PROP_SEARCH_PATH,
-    PROP_GC_NOTIFICATIONS
+    PROP_GC_NOTIFICATIONS,
+    PROP_PROGRAM_NAME,
 };
 
 
@@ -130,7 +133,7 @@ gjs_log(JSContext *context,
     JS_RestoreExceptionState(context, exc_state);
 
     if (jstr == NULL) {
-        gjs_debug(GJS_DEBUG_LOG, "<cannot convert value to string>");
+        g_message("JS LOG: <cannot convert value to string>");
         JS_EndRequest(context);
         return JS_TRUE;
     }
@@ -140,7 +143,7 @@ gjs_log(JSContext *context,
         return JS_FALSE;
     }
 
-    gjs_debug(GJS_DEBUG_LOG, "%s", s);
+    g_message("JS LOG: %s", s);
     g_free(s);
 
     JS_EndRequest(context);
@@ -154,43 +157,29 @@ gjs_log_error(JSContext *context,
               jsval     *vp)
 {
     jsval *argv = JS_ARGV(context, vp);
-    char *s;
     JSExceptionState *exc_state;
     JSString *jstr;
-    jsval exc;
-
-    if (argc != 2) {
-        gjs_throw(context, "Must pass an exception and message string to logError()");
-        return JS_FALSE;
-    }
 
     JS_BeginRequest(context);
 
-    exc = argv[0];
-
-    /* JS_ValueToString might throw, in which we will only
-     *log that the value could be converted to string */
-    exc_state = JS_SaveExceptionState(context);
-    jstr = JS_ValueToString(context, argv[1]);
-    if (jstr != NULL)
-        argv[1] = STRING_TO_JSVAL(jstr);    // GC root
-    JS_RestoreExceptionState(context, exc_state);
-
-    if (jstr == NULL) {
-        gjs_debug(GJS_DEBUG_ERROR, "<cannot convert value to string>");
-        gjs_log_exception_props(context, exc);
-        JS_EndRequest(context);
-        return JS_TRUE;
+    if ((argc != 1 && argc != 2) ||
+        !JSVAL_IS_OBJECT (argv[0])) {
+        gjs_throw(context, "Must pass an exception and optionally a message to logError()");
     }
 
-    if (!gjs_string_to_utf8(context, STRING_TO_JSVAL(jstr), &s)) {
-        JS_EndRequest(context);
-        return JS_FALSE;
+    if (argc == 2) {
+        /* JS_ValueToString might throw, in which we will only
+         *log that the value could be converted to string */
+        exc_state = JS_SaveExceptionState(context);
+        jstr = JS_ValueToString(context, argv[1]);
+        if (jstr != NULL)
+            argv[1] = STRING_TO_JSVAL(jstr);    // GC root
+        JS_RestoreExceptionState(context, exc_state);
+    } else {
+        jstr = NULL;
     }
 
-    gjs_debug(GJS_DEBUG_ERROR, "%s", s);
-    gjs_log_exception_props(context, exc);
-    g_free(s);
+    gjs_log_exception_full(context, argv[0], jstr);
 
     JS_EndRequest(context);
     JS_SET_RVAL(context, vp, JSVAL_VOID);
@@ -296,6 +285,8 @@ static void
 gjs_context_init(GjsContext *js_context)
 {
     js_context->jsversion_string = g_strdup(_GJS_JS_VERSION_DEFAULT);
+
+    gjs_context_make_current(js_context);
 }
 
 static void
@@ -341,6 +332,16 @@ gjs_context_class_init(GjsContextClass *klass)
                                     PROP_GC_NOTIFICATIONS,
                                     pspec);
 
+    pspec = g_param_spec_string("program-name",
+                                "Program Name",
+                                "The filename of the launched JS program",
+                                "",
+                                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
+
+    g_object_class_install_property(object_class,
+                                    PROP_PROGRAM_NAME,
+                                    pspec);
+
     signals[SIGNAL_GC] = g_signal_new("gc", G_TYPE_FROM_CLASS(klass),
                                       G_SIGNAL_RUN_LAST, 0,
                                       NULL, NULL,
@@ -384,13 +385,11 @@ gjs_context_dispose(GObject *object)
         gjs_debug(GJS_DEBUG_CONTEXT,
                   "Destroying JS context");
 
-        gjs_object_process_pending_toggles();
-
         /* Do a full GC here before tearing down, since once we do
          * that we may not have the JS_GetPrivate() to access the
          * context
          */
-        JS_GC(js_context->context);
+        JS_GC(js_context->runtime);
 
         gjs_object_process_pending_toggles();
 
@@ -399,6 +398,8 @@ gjs_context_dispose(GObject *object)
     }
 
     if (js_context->runtime != NULL) {
+        gjs_runtime_deinit(js_context->runtime);
+
         /* Cleans up data as well as destroying the runtime. */
         JS_DestroyRuntime(js_context->runtime);
         js_context->runtime = NULL;
@@ -426,6 +427,9 @@ gjs_context_finalize(GObject *object)
 
     g_free(js_context->jsversion_string);
 
+    if (gjs_context_get_current() == (GjsContext*)object)
+        gjs_context_make_current(NULL);
+
     g_mutex_lock(&contexts_lock);
     all_contexts = g_list_remove(all_contexts, object);
     g_mutex_unlock(&contexts_lock);
@@ -567,14 +571,14 @@ gjs_context_constructor (GType                  type,
 
     js_context->runtime = JS_NewRuntime(32*1024*1024 /* max bytes */);
     if (js_context->runtime == NULL)
-        gjs_fatal("Failed to create javascript runtime");
+        g_error("Failed to create javascript runtime");
     JS_SetGCParameter(js_context->runtime, JSGC_MAX_BYTES, 0xffffffff);
 
     js_context->context = JS_NewContext(js_context->runtime, 8192 /* stack chunk size */);
     if (js_context->context == NULL)
-        gjs_fatal("Failed to create javascript context");
+        g_error("Failed to create javascript context");
 
-    JS_SetRuntimePrivate(js_context->runtime, js_context->context);
+    gjs_runtime_init_for_context(js_context->runtime, js_context->context);
 
     JS_BeginRequest(js_context->context);
 
@@ -585,7 +589,7 @@ gjs_context_constructor (GType                  type,
      *
      * JSOPTION_STRICT: Report warnings to error reporter function.
      */
-    options_flags = JSOPTION_DONT_REPORT_UNCAUGHT | JSOPTION_STRICT;
+    options_flags = JSOPTION_DONT_REPORT_UNCAUGHT | JSOPTION_STRICT | JSOPTION_ALLOW_XML;
 
 #ifdef JSOPTION_JIT
     if (!g_getenv("GJS_DISABLE_JIT")) {
@@ -621,39 +625,39 @@ gjs_context_constructor (GType                  type,
     }
 
     if (!gjs_init_context_standard(js_context->context))
-        gjs_fatal("Failed to initialize context");
+        g_error("Failed to initialize context");
     js_context->global = JS_GetGlobalObject(js_context->context);
 
     if (!JS_DefineProperty(js_context->context, js_context->global,
                            "window", OBJECT_TO_JSVAL(js_context->global),
                            NULL, NULL,
                            JSPROP_READONLY | JSPROP_PERMANENT))
-        gjs_fatal("No memory to export global object as 'window'");
+        g_error("No memory to export global object as 'window'");
 
     /* Define a global function called log() */
     if (!JS_DefineFunction(js_context->context, js_context->global,
                            "log",
                            (JSNative)gjs_log,
                            1, GJS_MODULE_PROP_FLAGS))
-        gjs_fatal("Failed to define log function");
+        g_error("Failed to define log function");
 
     if (!JS_DefineFunction(js_context->context, js_context->global,
                            "logError",
                            (JSNative)gjs_log_error,
                            2, GJS_MODULE_PROP_FLAGS))
-        gjs_fatal("Failed to define logError function");
+        g_error("Failed to define logError function");
 
     /* Define global functions called print() and printerr() */
     if (!JS_DefineFunction(js_context->context, js_context->global,
                            "print",
                            (JSNative)gjs_print,
                            3, GJS_MODULE_PROP_FLAGS))
-        gjs_fatal("Failed to define print function");
+        g_error("Failed to define print function");
     if (!JS_DefineFunction(js_context->context, js_context->global,
                            "printerr",
                            (JSNative)gjs_printerr,
                            4, GJS_MODULE_PROP_FLAGS))
-        gjs_fatal("Failed to define printerr function");
+        g_error("Failed to define printerr function");
 
     /* We create the global-to-runtime root importer with the
      * passed-in search path. If someone else already created
@@ -664,19 +668,18 @@ gjs_context_constructor (GType                  type,
                                   (const char**) js_context->search_path :
                                   NULL,
                                   TRUE))
-        gjs_fatal("Failed to create root importer");
+        g_error("Failed to create root importer");
 
     /* Now copy the global root importer (which we just created,
      * if it didn't exist) to our global object
      */
     if (!gjs_define_root_importer(js_context->context,
-                                  js_context->global,
-                                  "imports"))
-        gjs_fatal("Failed to point 'imports' property at root importer");
+                                  js_context->global))
+        g_error("Failed to point 'imports' property at root importer");
 
     js_context->profiler = gjs_profiler_new(js_context->runtime);
 
-    JS_SetGCCallback(js_context->context, gjs_on_context_gc);
+    JS_SetGCCallback(js_context->runtime, gjs_on_context_gc);
 
     JS_EndRequest(js_context->context);
 
@@ -704,6 +707,9 @@ gjs_context_get_property (GObject     *object,
     case PROP_GC_NOTIFICATIONS:
         g_value_set_boolean(value, js_context->gc_notifications_enabled);
         break;
+    case PROP_PROGRAM_NAME:
+        g_value_set_string(value, js_context->program_name);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         break;
@@ -734,6 +740,9 @@ gjs_context_set_property (GObject      *object,
     case PROP_GC_NOTIFICATIONS:
         js_context->gc_notifications_enabled = g_value_get_boolean(value);
         break;
+    case PROP_PROGRAM_NAME:
+        js_context->program_name = g_value_dup_string(value);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         break;
@@ -881,7 +890,7 @@ gjs_context_maybe_gc (GjsContext  *context)
 void
 gjs_context_gc (GjsContext  *context)
 {
-    JS_GC(context->context);
+    JS_GC(context->runtime);
 }
 
 static gboolean
@@ -898,11 +907,12 @@ gjs_context_idle_emit_gc (gpointer data)
     return FALSE;
 }
 
-static JSBool
-gjs_on_context_gc (JSContext *cx,
+static void
+gjs_on_context_gc (JSRuntime *rt,
                    JSGCStatus status)
 {
-    GjsContext *gjs_context = JS_GetContextPrivate(cx);
+    JSContext *context = gjs_runtime_get_context(rt);
+    GjsContext *gjs_context = JS_GetContextPrivate(context);
 
     switch (status) {
         case JSGC_BEGIN:
@@ -921,8 +931,6 @@ gjs_on_context_gc (JSContext *cx,
         default:
         break;
     }
-
-    return TRUE;
 }
 
 /**
@@ -997,7 +1005,7 @@ gjs_context_eval(GjsContext *js_context,
     }
 
     /* log and clear exception if it's set (should not be, normally...) */
-    if (gjs_log_exception(js_context->context, NULL)) {
+    if (gjs_log_exception(js_context->context)) {
         gjs_debug(GJS_DEBUG_CONTEXT,
                   "Exception was set prior to JS_EvaluateScript()");
     }
@@ -1017,29 +1025,14 @@ gjs_context_eval(GjsContext *js_context,
                            filename,
                            line_number,
                            &retval)) {
-        char *message;
-
         gjs_debug(GJS_DEBUG_CONTEXT,
                   "Script evaluation failed");
 
-        /* if message is NULL then somehow exception wasn't set */
-        message = NULL;
-        gjs_log_exception(js_context->context,
-                             &message);
-        if (message) {
-            g_set_error(error,
-                        GJS_ERROR,
-                        GJS_ERROR_FAILED,
-                        "%s", message);
-            g_free(message);
-        } else {
-            gjs_debug(GJS_DEBUG_CONTEXT,
-                      "JS_EvaluateScript() failed but no exception message?");
-            g_set_error(error,
-                        GJS_ERROR,
-                        GJS_ERROR_FAILED,
-                        "JS_EvaluateScript() failed but no exception message?");
-        }
+        gjs_log_exception(js_context->context);
+        g_set_error(error,
+                    GJS_ERROR,
+                    GJS_ERROR_FAILED,
+                    "JS_EvaluateScript() failed");
 
         success = FALSE;
     }
@@ -1047,7 +1040,7 @@ gjs_context_eval(GjsContext *js_context,
     gjs_debug(GJS_DEBUG_CONTEXT,
               "Script evaluation succeeded");
 
-    if (gjs_log_exception(js_context->context, NULL)) {
+    if (gjs_log_exception(js_context->context)) {
         g_set_error(error,
                     GJS_ERROR,
                     GJS_ERROR_FAILED,
@@ -1110,26 +1103,29 @@ gjs_context_define_string_array(GjsContext  *js_context,
                                  js_context->global,
                                  array_name, array_length, array_values,
                                  JSPROP_READONLY | JSPROP_PERMANENT)) {
-        char *message;
-
-        message = NULL;
-        gjs_log_exception(js_context->context, &message);
-        if (message) {
-            g_set_error(error,
-                        GJS_ERROR,
-                        GJS_ERROR_FAILED,
-                        "%s", message);
-            g_free(message);
-        } else {
-            message = "gjs_define_string_array() failed but no exception message?";
-            gjs_debug(GJS_DEBUG_CONTEXT, "%s", message);
-            g_set_error(error,
-                        GJS_ERROR,
-                        GJS_ERROR_FAILED,
-                        "%s", message);
-        }
+        gjs_log_exception(js_context->context);
+        g_set_error(error,
+                    GJS_ERROR,
+                    GJS_ERROR_FAILED,
+                    "gjs_define_string_array() failed");
         return FALSE;
     }
 
     return TRUE;
 }
+
+static GjsContext *current_context;
+
+GjsContext *
+gjs_context_get_current (void)
+{
+    return current_context;
+}
+
+void
+gjs_context_make_current (GjsContext *context)
+{
+    g_assert (context == NULL || current_context == NULL);
+
+    current_context = context;
+}
