@@ -31,6 +31,7 @@
 #include "jsapi-util.h"
 #include "compat.h"
 #include "jsapi-private.h"
+#include "runtime.h"
 
 #include <string.h>
 #include <math.h>
@@ -52,10 +53,9 @@ gjs_new_object_for_constructor(JSContext *context,
     jsval     prototype;
 
     callee = JS_CALLEE(context, vp);
-    parent = JS_GetParent(context, JSVAL_TO_OBJECT (callee));
+    parent = JS_GetParent(JSVAL_TO_OBJECT (callee));
 
-    if (!gjs_object_get_property(context, JSVAL_TO_OBJECT (callee), "prototype",
-                                 &prototype))
+    if (!gjs_object_get_property_const(context, JSVAL_TO_OBJECT(callee), GJS_STRING_PROTOTYPE, &prototype))
         return NULL;
 
     return JS_NewObjectWithGivenProto(context, clasp,
@@ -184,7 +184,7 @@ gjs_typecheck_instance(JSContext *context,
 {
     if (!JS_InstanceOf(context, obj, static_clasp, NULL)) {
         if (throw) {
-            JSClass *obj_class = JS_GET_CLASS(context, obj);
+            JSClass *obj_class = JS_GetClass(obj);
 
             gjs_throw_custom(context, "TypeError",
                              "Object %p is not a subclass of %s, it's a %s",
@@ -206,11 +206,14 @@ gjs_construct_object_dynamic(JSContext      *context,
     JSObject *constructor;
     JSObject *result = NULL;
     jsval value;
+    jsid constructor_name;
 
     JS_BeginRequest(context);
 
+    constructor_name = gjs_runtime_get_const_string(JS_GetRuntime(context),
+                                                    GJS_STRING_CONSTRUCTOR);
     if (!gjs_object_require_property(context, proto, "prototype",
-                                     "constructor", &value))
+                                     constructor_name, &value))
         goto out;
 
     constructor = JSVAL_TO_OBJECT(value);
