@@ -31,6 +31,7 @@
 #include "jsapi-util.h"
 #include "compat.h"
 #include "jsapi-private.h"
+#include "runtime.h"
 
 #include <string.h>
 #include <math.h>
@@ -43,7 +44,6 @@ gjs_util_error_quark (void)
     return g_quark_from_static_string ("gjs-util-error-quark");
 }
 
-
 /**
  * gjs_get_import_global:
  * @context: a #JSContext
@@ -69,26 +69,33 @@ gjs_get_import_global(JSContext *context)
     return JS_GetGlobalObject(context);
 }
 
-/**
- * gjs_runtime_get_context:
- * @runtime: a #JSRuntime
- *
- * Gets the context associated with this runtime.
- *
- * Return value: the context, or %NULL if GJS hasn't been initialized
- * for the runtime or is being shut down.
- */
-JSContext *
-gjs_runtime_get_context(JSRuntime *runtime)
+static void
+global_finalize(JSFreeOp *fop, JSObject *global)
 {
-    return (JSContext *) JS_GetRuntimePrivate (runtime);
+    g_free(JS_GetPrivate(global));
+    JS_SetPrivate(global, NULL);
+}
+
+static void
+global_trace(JSTracer *trc, JSObject *global)
+{
+    jsval *slots;
+    int i;
+
+    slots = JS_GetPrivate(global);
+
+    for (i = 0; i < GJS_GLOBAL_SLOT_LAST; i++) {
+        if (!JSVAL_IS_VOID(slots[i]))
+            JS_CALL_VALUE_TRACER(trc, slots[i], "global slot");
+    }
 }
 
 static JSClass global_class = {
     "GjsGlobal", JSCLASS_GLOBAL_FLAGS,
     JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_StrictPropertyStub,
-    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
-    JSCLASS_NO_OPTIONAL_MEMBERS
+    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, global_finalize,
+    NULL /* checkAccess */, NULL /* call */, NULL /* hasInstance */, NULL /* construct */, global_trace,
+    { NULL }
 };
 
 /**
@@ -104,101 +111,96 @@ gboolean
 gjs_init_context_standard (JSContext       *context)
 {
     JSObject *global;
-    global = JS_NewCompartmentAndGlobalObject(context, &global_class, NULL);
+    jsval *slots;
+    int i;
+
+    global = JS_NewGlobalObject(context, &global_class, NULL);
     if (global == NULL)
         return FALSE;
     if (!JS_InitStandardClasses(context, global))
         return FALSE;
+
+    slots = g_new(jsval, GJS_GLOBAL_SLOT_LAST);
+    for (i = 0; i < GJS_GLOBAL_SLOT_LAST; i++)
+        slots[i] = JSVAL_VOID;
+
+    JS_SetPrivate(global, slots);
     return TRUE;
 }
 
-
-/* Checks whether an object has a property; unlike JS_GetProperty(),
- * never sets an exception. Treats a property with a value of JSVAL_VOID
- * the same as an absent property and returns false in both cases.
- */
-gboolean
-gjs_object_has_property(JSContext  *context,
-                        JSObject   *obj,
-                        const char *property_name)
+void
+gjs_set_global_slot (JSContext     *context,
+                     GjsGlobalSlot  slot,
+                     jsval          value)
 {
-    return gjs_object_get_property(context, obj, property_name, NULL);
+    JSObject *global;
+    jsval *slots;
+
+    global = JS_GetGlobalObject(context);
+    slots = JS_GetPrivate(global);
+    slots[slot] = value;
 }
 
-/* Checks whether an object has a property; unlike JS_GetProperty(),
- * never sets an exception. Treats a property with a value of JSVAL_VOID
- * the same as an absent property and returns false in both cases.
- * Always initializes *value_p, if only to JSVAL_VOID, even if it
- * returns FALSE.
- */
-gboolean
-gjs_object_get_property(JSContext  *context,
-                        JSObject   *obj,
-                        const char *property_name,
-                        jsval      *value_p)
+jsval
+gjs_get_global_slot (JSContext     *context,
+                     GjsGlobalSlot  slot)
 {
-    jsval value;
-    JSExceptionState *state;
+    JSObject *global;
+    jsval *slots;
 
-    JS_BeginRequest(context);
-
-    value = JSVAL_VOID;
-    state = JS_SaveExceptionState(context);
-    JS_GetProperty(context, obj, property_name, &value);
-    JS_RestoreExceptionState(context, state);
-
-    if (value_p)
-        *value_p = value;
-
-    JS_EndRequest(context);
-
-    return !JSVAL_IS_VOID(value);
+    global = JS_GetGlobalObject(context);
+    slots = JS_GetPrivate(global);
+    return slots[slot];
 }
 
 /* Returns whether the object had the property; if the object did
  * not have the property, always sets an exception. Treats
- * "the property's value is JSVAL_VOID" the same as "no such property,"
- * while JS_GetProperty() treats only "no such property" as an error.
+ * "the property's value is JSVAL_VOID" the same as "no such property,".
  * Guarantees that *value_p is set to something, if only JSVAL_VOID,
  * even if an exception is set and false is returned.
+ *
+ * Requires request.
  */
 gboolean
 gjs_object_require_property(JSContext       *context,
                             JSObject        *obj,
                             const char      *obj_description,
-                            const char      *property_name,
+                            jsid             property_name,
                             jsval           *value_p)
 {
     jsval value;
-
-    JS_BeginRequest(context);
+    char *name;
 
     value = JSVAL_VOID;
-    JS_GetProperty(context, obj, property_name, &value);
-
     if (value_p)
         *value_p = value;
 
-    if (!JSVAL_IS_VOID(value)) {
-        JS_ClearPendingException(context); /* in case JS_GetProperty() was on crack */
-        JS_EndRequest(context);
-        return TRUE;
-    } else {
-        /* remember gjs_throw() is a no-op if JS_GetProperty()
-         * already set an exception
-         */
-        if (obj_description)
-            gjs_throw(context,
-                      "No property '%s' in %s (or its value was undefined)",
-                      property_name, obj_description);
-        else
-            gjs_throw(context,
-                      "No property '%s' in object %p (or its value was undefined)",
-                      property_name, obj);
+    if (G_UNLIKELY (!JS_GetPropertyById(context, obj, property_name, &value)))
+        return JS_FALSE;
 
-        JS_EndRequest(context);
-        return FALSE;
+    if (G_LIKELY (!JSVAL_IS_VOID(value))) {
+        if (value_p)
+            *value_p = value;
+        return JS_TRUE;
     }
+
+    /* remember gjs_throw() is a no-op if JS_GetProperty()
+     * already set an exception
+     */
+
+    gjs_get_string_id(context, property_name, &name);
+
+    if (obj_description)
+        gjs_throw(context,
+                  "No property '%s' in %s (or its value was undefined)",
+                  name, obj_description);
+    else
+        gjs_throw(context,
+                  "No property '%s' in object %p (or its value was undefined)",
+                  name, obj);
+
+    g_free(name);
+    return JS_FALSE;
 }
 
 void
@@ -220,9 +222,9 @@ gjs_throw_abstract_constructor_error(JSContext *context,
     callee = JS_CALLEE(context, vp);
 
     if (JSVAL_IS_OBJECT(callee)) {
-        if (gjs_object_get_property(context, JSVAL_TO_OBJECT(callee),
-                                    "prototype", &prototype)) {
-            proto_class = JS_GetClass(context, JSVAL_TO_OBJECT(prototype));
+        if (gjs_object_get_property_const(context, JSVAL_TO_OBJECT(callee),
+                                          GJS_STRING_PROTOTYPE, &prototype)) {
+            proto_class = JS_GetClass(JSVAL_TO_OBJECT(prototype));
             name = proto_class->name;
         }
     }
@@ -352,7 +354,7 @@ gjs_value_debug_string(JSContext      *context,
              */
             JSClass *klass;
 
-            klass = JS_GET_CLASS(context, JSVAL_TO_OBJECT(value));
+            klass = JS_GetClass(JSVAL_TO_OBJECT(value));
             if (klass != NULL) {
                 str = JS_NewStringCopyZ(context, klass->name);
                 JS_ClearPendingException(context);
@@ -361,7 +363,7 @@ gjs_value_debug_string(JSContext      *context,
                     return g_strdup("[out of memory copying class name]");
                 }
             } else {
-                gjs_log_exception(context, NULL);
+                gjs_log_exception(context);
                 JS_EndRequest(context);
                 return g_strdup("[unknown object]");
             }
@@ -408,8 +410,7 @@ gjs_log_object_props(JSContext      *context,
 
     props_iter = JS_NewPropertyIterator(context, obj);
     if (props_iter == NULL) {
-        gjs_debug(GJS_DEBUG_ERROR,
-                  "Failed to create property iterator for object props");
+        gjs_log_exception(context);
         goto done;
     }
 
@@ -419,13 +420,13 @@ gjs_log_object_props(JSContext      *context,
 
     while (!JSID_IS_VOID(prop_id)) {
         jsval propval;
-        char *name;
         char *debugstr;
+        char *name;
 
-        if (!gjs_get_string_id(context, prop_id, &name))
+        if (!JS_GetPropertyById(context, obj, prop_id, &propval))
             goto next;
 
-        if (!gjs_object_get_property(context, obj, name, &propval))
+        if (!gjs_get_string_id(context, prop_id, &name))
             goto next;
 
         debugstr = gjs_value_debug_string(context, propval);
@@ -488,7 +489,7 @@ gjs_explain_scope(JSContext  *context,
         g_string_append_printf(chain, "%p %s",
                                parent, debug);
         g_free(debug);
-        parent = JS_GetParent(context, parent);
+        parent = JS_GetParent(parent);
     }
     gjs_debug(GJS_DEBUG_SCOPE,
               "  Chain: %s",
@@ -500,99 +501,71 @@ gjs_explain_scope(JSContext  *context,
     JS_EndRequest(context);
 }
 
-static void
-log_one_exception_property(JSContext  *context,
-                           JSObject   *object,
-                           const char *name)
+JSBool
+gjs_log_exception_full(JSContext *context,
+                       jsval      exc,
+                       JSString  *message)
 {
-    jsval v;
-    char *debugstr;
+    jsval stack;
+    JSString *exc_str;
+    char *utf8_exception, *utf8_message, *utf8_stack;
 
-    gjs_object_get_property(context, object, name, &v);
-
-    debugstr = gjs_value_debug_string(context, v);
-    gjs_debug(GJS_DEBUG_ERROR, "  %s = '%s'", name, debugstr);
-    g_free(debugstr);
-}
-
-void
-gjs_log_exception_props(JSContext *context,
-                        jsval      exc)
-{
     JS_BeginRequest(context);
 
-    /* This is useful when the exception was never sent to an error reporter
-     * due to JSOPTION_DONT_REPORT_UNCAUGHT, or if the exception was not
-     * a normal Error object so jsapi didn't know how to report it sensibly.
-     */
-    if (JSVAL_IS_NULL(exc)) {
-        gjs_debug(GJS_DEBUG_ERROR,
-                  "Exception was null");
-    } else if (JSVAL_IS_OBJECT(exc)) {
-        JSObject *exc_obj;
+    exc_str = JS_ValueToString(context, exc);
+    if (exc_str != NULL)
+        gjs_string_to_utf8(context, STRING_TO_JSVAL(exc_str), &utf8_exception);
+    else
+        utf8_exception = NULL;
 
-        exc_obj = JSVAL_TO_OBJECT(exc);
+    if (message != NULL)
+        gjs_string_to_utf8(context, STRING_TO_JSVAL(message), &utf8_message);
+    else
+        utf8_message = NULL;
 
-        log_one_exception_property(context, exc_obj, "message");
-        log_one_exception_property(context, exc_obj, "fileName");
-        log_one_exception_property(context, exc_obj, "lineNumber");
-        log_one_exception_property(context, exc_obj, "stack");
-    } else if (JSVAL_IS_STRING(exc)) {
-        gjs_debug(GJS_DEBUG_ERROR,
-                  "Exception was a String");
+    if (JSVAL_IS_OBJECT(exc) &&
+        gjs_object_get_property_const(context, JSVAL_TO_OBJECT(exc),
+                                      GJS_STRING_STACK, &stack) &&
+        JSVAL_IS_STRING(stack))
+        gjs_string_to_utf8(context, stack, &utf8_stack);
+    else
+        utf8_stack = NULL;
+
+    if (utf8_message) {
+        if (utf8_stack)
+            g_warning("JS ERROR: %s: %s\n%s", utf8_message, utf8_exception, utf8_stack);
+        else
+            g_warning("JS ERROR: %s: %s", utf8_message, utf8_exception);
     } else {
-        gjs_debug(GJS_DEBUG_ERROR,
-                  "Exception had some strange type");
+        if (utf8_stack)
+            g_warning("JS ERROR: %s\n%s", utf8_exception, utf8_stack);
+        else
+            g_warning("JS ERROR: %s", utf8_exception);
     }
-    JS_EndRequest(context);
+
+    g_free(utf8_exception);
+    g_free(utf8_message);
+    g_free(utf8_stack);
+
+    return JS_TRUE;
 }
 
 static JSBool
 log_and_maybe_keep_exception(JSContext  *context,
-                             char      **message_p,
                              gboolean    keep)
 {
     jsval exc = JSVAL_VOID;
-    JSString *s;
-    char *message;
     JSBool retval = JS_FALSE;
 
     JS_BeginRequest(context);
 
-    if (message_p)
-        *message_p = NULL;
-
     JS_AddValueRoot(context, &exc);
     if (!JS_GetPendingException(context, &exc))
         goto out;
 
     JS_ClearPendingException(context);
 
-    s = JS_ValueToString(context, exc);
-
-    if (s == NULL) {
-        gjs_debug(GJS_DEBUG_ERROR,
-                  "Failed to convert exception to string");
-        goto out; /* Exception should be thrown already */
-    }
-
-    if (!gjs_string_to_utf8(context, STRING_TO_JSVAL(s), &message)) {
-        gjs_debug(GJS_DEBUG_ERROR,
-                  "Failed to convert exception string to UTF-8");
-        goto out; /* Error already set */
-    }
-
-    gjs_debug(GJS_DEBUG_ERROR,
-              "Exception was: %s",
-              message);
-
-    if (message_p) {
-        *message_p = message;
-    } else {
-        g_free(message);
-    }
-
-    gjs_log_exception_props(context, exc);
+    gjs_log_exception_full(context, exc, NULL);
 
     /* We clear above and then set it back so any exceptions
      * from the logging process don't overwrite the original
@@ -611,17 +584,15 @@ log_and_maybe_keep_exception(JSContext  *context,
 }
 
 JSBool
-gjs_log_exception(JSContext  *context,
-                  char      **message_p)
+gjs_log_exception(JSContext  *context)
 {
-    return log_and_maybe_keep_exception(context, message_p, FALSE);
+    return log_and_maybe_keep_exception(context, FALSE);
 }
 
 JSBool
-gjs_log_and_keep_exception(JSContext *context,
-                           char     **message_p)
+gjs_log_and_keep_exception(JSContext *context)
 {
-    return log_and_maybe_keep_exception(context, message_p, TRUE);
+    return log_and_maybe_keep_exception(context, TRUE);
 }
 
 static void
@@ -647,12 +618,12 @@ try_to_chain_stack_trace(JSContext *src_context, JSContext *dst_context,
     JS_ClearPendingException(dst_context);
 
     /* get stack trace for src_exc and chained */
-    if (!(gjs_object_get_property(dst_context, JSVAL_TO_OBJECT(chained),
-                                  "stack", &dst_stack) &&
+    if (!(JS_GetProperty(dst_context, JSVAL_TO_OBJECT(chained),
+                         "stack", &dst_stack) &&
           JSVAL_IS_STRING(dst_stack)))
         goto out; // couldn't get chained stack
-    if (!(gjs_object_get_property(src_context, JSVAL_TO_OBJECT(src_exc),
-                                  "stack", &src_stack) &&
+    if (!(JS_GetProperty(src_context, JSVAL_TO_OBJECT(src_exc),
+                         "stack", &src_stack) &&
           JSVAL_IS_STRING(src_stack)))
         goto out; // couldn't get source stack
 
@@ -810,7 +781,6 @@ jsval
 gjs_date_from_time_t (JSContext *context, time_t time)
 {
     JSObject *date;
-    JSClass *date_class;
     JSObject *date_constructor;
     jsval date_prototype;
     jsval args[1];
@@ -825,18 +795,15 @@ gjs_date_from_time_t (JSContext *context, time_t time)
 
     if (!JS_GetClassObject(context, JS_GetGlobalObject(context), JSProto_Date,
                            &date_constructor))
-        gjs_fatal("Failed to lookup Date prototype");
+        g_error("Failed to lookup Date prototype");
 
     if (!JS_GetProperty(context, date_constructor, "prototype", &date_prototype))
-        gjs_fatal("Failed to get prototype from Date constructor");
-
-    date_class = JS_GET_CLASS(context, JSVAL_TO_OBJECT (date_prototype));
+        g_error("Failed to get prototype from Date constructor");
 
     if (!JS_NewNumberValue(context, ((double) time) * 1000, &(args[0])))
-        gjs_fatal("Failed to convert time_t to number");
+        g_error("Failed to convert time_t to number");
 
-    date = JS_ConstructObjectWithArguments(context, date_class,
-                                           NULL, NULL, 1, args);
+    date = JS_New(context, JSVAL_TO_OBJECT (date_prototype), 1, args);
 
     result = OBJECT_TO_JSVAL(date);
     JS_LeaveLocalRootScope(context);
@@ -1198,7 +1165,7 @@ gjs_maybe_gc (JSContext *context)
          */
         if (rss_size > linux_rss_trigger) {
             linux_rss_trigger = (gulong) MIN(G_MAXULONG, rss_size * 1.25);
-            JS_GC(context);
+            JS_GC(JS_GetRuntime(context));
         } else if (rss_size < (0.75 * linux_rss_trigger)) {
             /* If we've shrunk by 75%, lower the trigger */
             linux_rss_trigger = (rss_size * 1.25);
