@@ -102,7 +102,7 @@ gjs_gc(JSContext *context,
     jsval *argv = JS_ARGV(cx, vp);
     if (!gjs_parse_args(context, "gc", "", argc, argv))
         return JS_FALSE;
-    JS_GC(context);
+    JS_GC(JS_GetRuntime(context));
     return JS_TRUE;
 }
 
@@ -123,6 +123,11 @@ JSBool
 gjs_js_define_system_stuff(JSContext *context,
                            JSObject  *module)
 {
+    GjsContext *gjs_context;
+    char *program_name;
+    jsval value;
+    JSBool retval;
+
     if (!JS_DefineFunction(context, module,
                            "addressOf",
                            (JSNative) gjs_address_of,
@@ -153,5 +158,39 @@ gjs_js_define_system_stuff(JSContext *context,
                            0, GJS_MODULE_PROP_FLAGS))
         return JS_FALSE;
 
-    return JS_TRUE;
+    retval = JS_FALSE;
+
+    gjs_context = JS_GetContextPrivate(context);
+    g_object_get(gjs_context,
+                 "program-name", &program_name,
+                 NULL);
+
+    if (!gjs_string_from_utf8(context, program_name,
+                              -1, &value))
+        goto out;
+
+    /* The name is modeled after program_invocation_name,
+       part of the glibc */
+    if (!JS_DefineProperty(context, module,
+                           "programInvocationName",
+                           value,
+                           JS_PropertyStub,
+                           JS_StrictPropertyStub,
+                           GJS_MODULE_PROP_FLAGS | JSPROP_READONLY))
+        goto out;
+
+    if (!JS_DefineProperty(context, module,
+                           "version",
+                           INT_TO_JSVAL(GJS_VERSION),
+                           JS_PropertyStub,
+                           JS_StrictPropertyStub,
+                           GJS_MODULE_PROP_FLAGS | JSPROP_READONLY))
+        goto out;
+
+    retval = JS_TRUE;
+
+ out:
+    g_free(program_name);
+
+    return retval;
 }
